/**********************************************************************
*
COPYRIGHT (C) 2000 Thomas Becker

PERMISSION NOTICE:

PERMISSION TO USE, COPY, MODIFY, AND DISTRIBUTE THIS SOFTWARE FOR ANY
PURPOSE AND WITHOUT FEE IS HEREBY GRANTED, PROVIDED THAT ALL COPIES
ARE ACCOMPANIED BY THE COMPLETE MACHINE-READABLE SOURCE CODE, ALL
MODIFIED FILES CARRY PROMINENT NOTICES AS TO WHEN AND BY WHOM THEY
WERE MODIFIED, THE ABOVE COPYRIGHT NOTICE, THIS PERMISSION NOTICE AND
THE NO-WARRANTY NOTICE BELOW APPEAR IN ALL SOURCE FILES AND IN ALL
SUPPORTING DOCUMENTATION.

NO-WARRANTY NOTICE:

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
MERCHANTABILITY ARE HEREBY DISCLAIMED.
*
**********************************************************************/
 
//
// File: FairMonitor.cpp
//
// Author: Thomas Becker
//
// Created: Dec 2000
//
// Description: Implementation of class FairMonitor
//

#include"stdafx.h"
#include "FairMonitor.h"

///////////////////////////////////////////////////////////////////////
//
FairMonitor::FairMonitor() :
  m_nLockCount(0)
{
  // Initialize the critical section that protects access to
  // the wait set. There is no way to check for errors here.
  ::InitializeCriticalSection(&m_critsecWaitSetProtection);

  // Initialize the critical section that provides synchronization
  // to the client. There is no way to check for errors here.
  ::InitializeCriticalSection(&m_critsecSynchronized);
}

///////////////////////////////////////////////////////////////////////
//
FairMonitor::~FairMonitor()
{ Cleanup(); }

///////////////////////////////////////////////////////////////////////
//
void FairMonitor::BeginSynchronized()
{ 
  // Acquire lock. Win32 allows no error checking here.
  ::EnterCriticalSection(&m_critsecSynchronized); 

  // Record the lock acquisition for proper release in Wait().
  ++m_nLockCount;
}

///////////////////////////////////////////////////////////////////////
//
void FairMonitor::EndSynchronized()
{ 
  // Record the lock release for proper release in Wait().
  --m_nLockCount;

  // Release lock. Win32 allows no error checking here.
  ::LeaveCriticalSection(&m_critsecSynchronized); 

}

///////////////////////////////////////////////////////////////////////
//
DWORD FairMonitor::Wait(
  DWORD dwMillisecondsTimeout/* = INFINITE*/,
  BOOL bAlertable/* = FALSE */
  )
{
  // Enter a new event handle into the wait set.
  HANDLE hWaitEvent = Push();
  if( NULL == hWaitEvent )
    return WAIT_FAILED;

  // Store the current lock count for re-acquisition.
  int nThisThreadsLockCount = m_nLockCount;
  m_nLockCount = 0;

  // Release the synchronization lock the appropriate number of times.
  // Win32 allows no error checking here.
  for( int i=0; i<nThisThreadsLockCount; ++i)
    ::LeaveCriticalSection(&m_critsecSynchronized);

  // NOTE: We're now in the wait set, so if we get preempted here,
  // that's fine. If we're getting notified during this gap, the
  // wait operation below will find the event signalled.
  
  // Wait for the event to become signalled.
  DWORD dwWaitResult = ::WaitForSingleObjectEx(
    hWaitEvent,
    dwMillisecondsTimeout,
    bAlertable
    );

  // If the wait failed, store the last error because it will get
  // overwritten when acquiring the lock.
  DWORD dwLastError;
  if( WAIT_FAILED == dwWaitResult )
    dwLastError = ::GetLastError();

  // Acquire the synchronization lock the appropriate number of times.
  // Win32 allows no error checking here.
  for( int j=0; j<nThisThreadsLockCount; ++j)
    ::EnterCriticalSection(&m_critsecSynchronized);

  // Restore lock count.
  m_nLockCount = nThisThreadsLockCount;

  // Close event handle
  if( ! CloseHandle(hWaitEvent) )
    return WAIT_FAILED;

  if( WAIT_FAILED == dwWaitResult )
    ::SetLastError(dwLastError);

  return dwWaitResult;
}

///////////////////////////////////////////////////////////////////////
//
BOOL FairMonitor::Notify()
{
  // Pop the first handle, if any, off the wait set.
  HANDLE hWaitEvent = Pop();

  // If there is not thread currently waiting, that's just fine.
  if(NULL == hWaitEvent)
    return TRUE;

  // Signal the event.
  return SetEvent(hWaitEvent);
}

///////////////////////////////////////////////////////////////////////
//
BOOL FairMonitor::NotifyAll()
{
  // Signal all events on the deque, then clear it. Win32 allows no
  // error checking on entering and leaving the critical section.
  //
  ::EnterCriticalSection(&m_critsecWaitSetProtection);
  std::deque<HANDLE>::const_iterator it_run = m_deqWaitSet.begin();
  std::deque<HANDLE>::const_iterator it_end = m_deqWaitSet.end();
  for( ; it_run < it_end; ++it_run )
  {
    if( ! SetEvent(*it_run) )
      return FALSE;
  }
  m_deqWaitSet.clear();
  ::LeaveCriticalSection(&m_critsecWaitSetProtection);
  
  return TRUE;
}

///////////////////////////////////////////////////////////////////////
//
void FairMonitor::Cleanup()
{
  // Uninitialize critical sections. Win32 allows no error checking
  // here.
  ::DeleteCriticalSection(&m_critsecWaitSetProtection);
  ::DeleteCriticalSection(&m_critsecSynchronized);

  // Free wait set memory (really, free it!).
  m_deqWaitSet.swap(std::deque<HANDLE>());
}

///////////////////////////////////////////////////////////////////////
//
HANDLE FairMonitor::Push()
{
  // Create the new event.
  HANDLE hWaitEvent = ::CreateEvent(
    NULL, // no security
    FALSE, // auto-reset event
    FALSE, // initially unsignalled
    NULL // string name
    );
  //
  if( NULL == hWaitEvent ) {
    return NULL;
  }

  // Push the handle on the deque.
  ::EnterCriticalSection(&m_critsecWaitSetProtection);
  m_deqWaitSet.push_back(hWaitEvent);
  ::LeaveCriticalSection(&m_critsecWaitSetProtection);

  return hWaitEvent;
}

///////////////////////////////////////////////////////////////////////
//
HANDLE FairMonitor::Pop()
{
  // Pop the first handle off the deque.
  //
  ::EnterCriticalSection(&m_critsecWaitSetProtection);
  HANDLE hWaitEvent = NULL; 
  if( 0 != m_deqWaitSet.size() )
  {
    hWaitEvent = m_deqWaitSet.front();
    m_deqWaitSet.pop_front();
  }
  ::LeaveCriticalSection(&m_critsecWaitSetProtection);

  return hWaitEvent;
}
