/**********************************************************************
*
COPYRIGHT (C) 2000 Thomas Becker

PERMISSION NOTICE:

PERMISSION TO USE, COPY, MODIFY, AND DISTRIBUTE THIS SOFTWARE FOR ANY
PURPOSE AND WITHOUT FEE IS HEREBY GRANTED, PROVIDED THAT ALL COPIES
ARE ACCOMPANIED BY THE COMPLETE MACHINE-READABLE SOURCE CODE, ALL
MODIFIED FILES CARRY PROMINENT NOTICES AS TO WHEN AND BY WHOM THEY
WERE MODIFIED, THE ABOVE COPYRIGHT NOTICE, THIS PERMISSION NOTICE AND
THE NO-WARRANTY NOTICE BELOW APPEAR IN ALL SOURCE FILES AND IN ALL
SUPPORTING DOCUMENTATION.

NO-WARRANTY NOTICE:

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
MERCHANTABILITY ARE HEREBY DISCLAIMED.
*
**********************************************************************/

 
//
// File: FairMonitor.h
//
// Author: Thomas Becker
//
// Created: Dec 2000
//
// Description: Declaration of class FairMonitor
//

// Multiple inclusion protection
//
#ifndef __FAIR_MONITOR_INCLUDED__
#define __FAIR_MONITOR_INCLUDED__

#include"stdafx.h"
#include<deque>

///////////////////////////////////////////////////////////////////////
//
// The class FairMonitor implements a Java-style monitor in Win32. This
// class employs a significant amount of user-mode code in its Wait(), 
// Notify(), and NotifyAll() functions. If it is known beforehand that 
// only one of Notify() and NotifyAll() will be needed, using the class
// NotifyMonitor or NotifyAllMonitor instead is more efficient.
//
// The name FairMonitor was chosen because the class guarantees that
// the Wait() and Notfiy() methods treat the wait set as a FIFO queue.
//
// This is the only one of the Win32 monitor classes that is supported
// by Windows 95/98.
//
class FairMonitor
{
public:

  // Public Construction and Destruction
  // ===================================
  
  // Default constructor.
  FairMonitor();

  // The destructor closes the handles and frees the memory
  // for the wait set. NOTE: If you must derive from this class,
  // you should make the destructor virtual.
  ~FairMonitor();

  // Operations
  // ==========
  
  // Acquires the lock once. There is no error checking as the lock is
  // implemented as a Win32 critical section.
  void BeginSynchronized();

  // Releases the lock once. As the lock is implemented as a Win32 
  // critical section, no error checking is possible. In particular,
  // it is the client's responsibility to make sure that only threads
  // that are currently in possession of the lock call this function.
  void EndSynchronized();

  // Waits for a notification. The first argument is the timeout
  // in milliseconds. The alertable flag must be true if the wait
  // should complete when a user APC is queued.
  //
  // In case of success, the return value is WAIT_OBJECT_0. See
  // the documentation of the Win32 function WaitForSingleObjectEx()
  // for an explanation of the other possible return values.
  //
  // NOTE: The return value WAIT_FAILED indicates an irrecoverably bad
  // state of the process, usually caused by a bad API call such as 
  // entering or leaving a destroyed critical section.
  //
  // NOTE: Since the lock is implemented as a Win32 critical section,
  // there is no way to check whether or not the lock has been
  // acquired before this function gets called. IT IS THE 
  // RESPONSIBILITY OF THE CLIENT TO MAKE SURE THAT THIS FUNCTION IS
  // NOT CALLED BY ANY THREAD THAT DOES NOT CURRENTLY OWN THE MONITOR'S
  // LOCK. TERRIBLE THINGS WILL HAPPEN IF THIS RULE IS VIOLATED.
  DWORD Wait(
    DWORD dwMillisecondsTimeout = INFINITE,
    BOOL bAlertable = FALSE
    );

  // Notifies one of the waiting threads, if any. In case of success, 
  // the return value is TRUE. In case of failure, call GetLastError()
  // to obtain error information.
  //
  // NOTE: The return value FALSE indicates an irrecoverrably bad state
  // of the process, usually caused by a bad API call such entering or
  // leaving a destroyed critical section.
  BOOL Notify();

  // Notifies the waiting threads, if any. In case of success, the 
  // return value is TRUE. In case of failure, call GetLastError() to 
  // obtain error information.
  //
  // NOTE: The return value FALSE indicates an irrecoverrably bad state
  // of the process, usually caused by a bad API call such entering or
  // leaving a destroyed critical section.
  BOOL NotifyAll();

  // Checks if the handles to the synch objects are non-NULL.
  bool IsValid();

private:

  // Private helper functions
  // ========================
  
  // Uninitializes the sync objects and empties the wait set.
  void Cleanup();

  // Creates an initially non-signalled auto-reset event and
  // pushes the handle to the event onto the wait set. The
  // return value is the event handle. In case of failure,
  // NULL is returned.
  HANDLE Push();

  // Pops the first handle off the wait set. Returns NULL if the
  // wait set was empty.
  HANDLE Pop();

  // Data members
  // ============

  // STL deque that implements the wait set.
  std::deque<HANDLE> m_deqWaitSet;

  // Critical section to protect access to wait set.
  CRITICAL_SECTION m_critsecWaitSetProtection;

  // Critical section for external synchronization.
  CRITICAL_SECTION m_critsecSynchronized;

  // The monitor must keep track of how many times the lock
  // has been acquired, because Win32 does not divulge this
  // information to the client programmer.
  int m_nLockCount;

};

#endif
//
// End multiple inclusion protection
