/**********************************************************************
*
COPYRIGHT (C) 2000 Thomas Becker

PERMISSION NOTICE:

PERMISSION TO USE, COPY, MODIFY, AND DISTRIBUTE THIS SOFTWARE FOR ANY
PURPOSE AND WITHOUT FEE IS HEREBY GRANTED, PROVIDED THAT ALL COPIES
ARE ACCOMPANIED BY THE COMPLETE MACHINE-READABLE SOURCE CODE, ALL
MODIFIED FILES CARRY PROMINENT NOTICES AS TO WHEN AND BY WHOM THEY
WERE MODIFIED, THE ABOVE COPYRIGHT NOTICE, THIS PERMISSION NOTICE AND
THE NO-WARRANTY NOTICE BELOW APPEAR IN ALL SOURCE FILES AND IN ALL
SUPPORTING DOCUMENTATION.

NO-WARRANTY NOTICE:

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
MERCHANTABILITY ARE HEREBY DISCLAIMED.
*
**********************************************************************/
 
//
// File: MonitorImpl.cpp
//
// Author: Thomas Becker
//
// Created: Dec 2000
//
// Description: Implementation of class MonitorImpl
//

#include "stdafx.h"
#include "MonitorImpl.h"

///////////////////////////////////////////////////////////////////////
//
MonitorImpl::MonitorImpl() : 
  m_hNotificationEvent(NULL),
  m_hSynchronizationMutex(NULL),
  m_nLockCount(0)
  {/* All initialization takes place in the Create() method. */}

///////////////////////////////////////////////////////////////////////
//
MonitorImpl::~MonitorImpl()
{ Cleanup(); }

///////////////////////////////////////////////////////////////////////
//
bool MonitorImpl::Create(NotificationType enumType)
{
  // Check if already created.
  if( IsValid() ) {
    ::SetLastError(ERROR_ALREADY_EXISTS);
    return false;
  }
  
  // Create the notification event.
  m_hNotificationEvent = ::CreateEvent(
    NULL, // no security
    enumType == enNotify ? FALSE : TRUE, // auto-reset event
    FALSE, // initially unsignalled
    NULL // string name
    );
  //
  if( NULL == m_hNotificationEvent ) {
    // Store last error as it may be overwritten by Cleanup().
    DWORD dwLastError = GetLastError();
    Cleanup();
    // Restore last error.
    SetLastError(dwLastError);
    return false;
  }

  // Create the synchronization mutex.
  m_hSynchronizationMutex = ::CreateMutex(
    NULL, // no security
    FALSE, // initially not owned
    NULL // string name
    );
  //
  if( NULL == m_hSynchronizationMutex ) {
    // Store last error as it may be overwritten by Cleanup().
    DWORD dwLastError = GetLastError();
    Cleanup();
    // Restore last error.
    SetLastError(dwLastError);
    return false;
  }

  return true;
}

///////////////////////////////////////////////////////////////////////
//
DWORD MonitorImpl::BeginSynchronized()
{
  // Wait for mutex.
  DWORD dwWaitRetVal = ::WaitForSingleObject(
    m_hSynchronizationMutex, 
    INFINITE
    );

  // If the wait completed successfully, record the lock acquisition 
  // for proper release in Wait().
  if( WAIT_OBJECT_0 == dwWaitRetVal )
    ++m_nLockCount;

  return dwWaitRetVal;
}

///////////////////////////////////////////////////////////////////////
//
BOOL MonitorImpl::EndSynchronized()
{
  // Record the lock release for proper release in Wait().
  --m_nLockCount;

  // Release mutex.
  BOOL bReleaseRetVal = ::ReleaseMutex(m_hSynchronizationMutex);

  return bReleaseRetVal;
}

///////////////////////////////////////////////////////////////////////
//
DWORD MonitorImpl::Wait(
  DWORD dwMillisecondsTimeout,
  BOOL bAlertable
  )
{
  // Store the current lock count for re-acquisition.
  int nThisThreadsLockCount = m_nLockCount;
  m_nLockCount = 0;

  // Release the mutex enough times so that it is owned exactly once.
  //
  for( int i=0; i<nThisThreadsLockCount-1; ++i )
  {
    BOOL bReleaseRetVal = ::ReleaseMutex(m_hSynchronizationMutex);
    // If the release failed, we must return WAIT_FAILED. The client
    // will call GetLastError() to find out what really happened.
    if( ! bReleaseRetVal )
      return WAIT_FAILED;
  }
  
  DWORD dwWaitResult = ::SignalObjectAndWait(
    m_hSynchronizationMutex,
    m_hNotificationEvent,
    dwMillisecondsTimeout,
    bAlertable
    );

  // If the wait failed, store last error. And YES, the documentation
  // specifies 0xFFFFFFFF, not WAIT_FAILED as the return value on 
  // failure.
  DWORD dwWaitError;
  if( 0xFFFFFFFF == dwWaitResult )
  {
    dwWaitResult = WAIT_FAILED; // That's what it should have been.
    dwWaitError = ::GetLastError();
  }

  // Re-acquire the lock. NOTE: Do not use an alertable wait
  // here! The lock must be reacquired when the wait is 
  // interrupted
  //
  DWORD dwLockResult;
  for( int j=0; j<nThisThreadsLockCount; ++j )
  {
    dwLockResult = ::WaitForSingleObject(
      m_hSynchronizationMutex,
      INFINITE
      );
    if( WAIT_OBJECT_0 != dwLockResult )
      break;
  }

  // If either one of the wait operations failed, we're irrecoverably
  // screwed.
  if( WAIT_FAILED == dwWaitResult || 
      WAIT_FAILED == dwLockResult ||
      WAIT_ABANDONED == dwLockResult )
  {
    // If the second wait was successful, we must restore the error
    // code for the client.
    if( WAIT_OBJECT_0 == dwLockResult )
    {
      m_nLockCount = nThisThreadsLockCount;
      ::SetLastError(dwWaitError);
      return dwWaitResult;
    }

    return dwLockResult;
  }

  // What matters to the caller is the return value of the wait.
  m_nLockCount = nThisThreadsLockCount;
  return dwWaitResult;

}

///////////////////////////////////////////////////////////////////////
//
BOOL MonitorImpl::Pulse()
{
  return ::PulseEvent(m_hNotificationEvent);
}

///////////////////////////////////////////////////////////////////////
//
void MonitorImpl::Cleanup()
{
  //
  // NOTE: This function is called only in the destructor and when
  // some Win32 call has failed. Therefore, we do not check the
  // return value of CloseHandle() here.
  //

  // Close all handles.
  if(NULL != m_hNotificationEvent) {
    CloseHandle(m_hNotificationEvent);
    m_hNotificationEvent = NULL;
  }
  if(NULL != m_hSynchronizationMutex) {
    CloseHandle(m_hSynchronizationMutex);
    m_hSynchronizationMutex = NULL;
  }
}

///////////////////////////////////////////////////////////////////////
//
bool MonitorImpl::IsValid()
{
  return NULL != m_hNotificationEvent && NULL != m_hSynchronizationMutex;
}

