/**********************************************************************
*
COPYRIGHT (C) 2000 Thomas Becker

PERMISSION NOTICE:

PERMISSION TO USE, COPY, MODIFY, AND DISTRIBUTE THIS SOFTWARE FOR ANY
PURPOSE AND WITHOUT FEE IS HEREBY GRANTED, PROVIDED THAT ALL COPIES
ARE ACCOMPANIED BY THE COMPLETE MACHINE-READABLE SOURCE CODE, ALL
MODIFIED FILES CARRY PROMINENT NOTICES AS TO WHEN AND BY WHOM THEY
WERE MODIFIED, THE ABOVE COPYRIGHT NOTICE, THIS PERMISSION NOTICE AND
THE NO-WARRANTY NOTICE BELOW APPEAR IN ALL SOURCE FILES AND IN ALL
SUPPORTING DOCUMENTATION.

NO-WARRANTY NOTICE:

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
MERCHANTABILITY ARE HEREBY DISCLAIMED.
*
**********************************************************************/
 
//
// File: MonitorImpl.h
//
// Author: Thomas Becker
//
// Created: Dec 2000
//
// Description: Declaration of class MonitorImpl
//

// Multiple inclusion protection
//
#ifndef __MONITOR_IMPL_INCLUDED__
#define __MONITOR_IMPL_INCLUDED__

#include"stdafx.h"

///////////////////////////////////////////////////////////////////////
//
// The class MonitorImpl implements a Java-style monitor in Win32. 
// Any one object of this type can only support Notify() or NotifyAll().
// The type of notification supported is specified as a creation flag.
//
// This class internally uses the Win32 function SignalObjectAndWait(),
// which is supported only under Windows NT 4.0 and higher. Therefore,
// you must define _WIN32_WINNT 0x0400 for this to work.
//
class MonitorImpl
{
public:

  // Enums
  // =====
  enum NotificationType
  {
    enNotify,
    enNotifyAll
  };
  
  // Public Construction and Destruction
  // ===================================
  
  // Default constructor. The MonitorImpl is not valid until the
  // Create() member function has been called.
  MonitorImpl();

  // The destructor closes all handles.
  ~MonitorImpl();

  // Creates all synchronization objects. Returns true in case of
  // success, false otherwise. In case of failure, call GetLastError()
  // to obtain error information.
  bool Create(NotificationType enumType);

  // Operations
  // ==========
  
  // Acquires the lock once. In case of sucess, the return value is
  // WAIT_OBJECT_0. See the documentation of the Win32 function
  // WaitForSingleObject() for an explanation of the other possible
  // return values. The object being waited on is a mutex. 
  //
  // NOTE: The return value WAIT_FAILED indicates an irrecoverably
  // bad state of the process, usually caused by a bad API call such as
  // entering or leaving a destroyed critical section.
  DWORD BeginSynchronized();

  // Releases the lock once. In case of success, the return value is 
  // TRUE. In case of failure, call GetLastError() to obtain error
  // information.
  //
  // NOTE: The return value FALSE indicates an irrecoverably bad state
  // of the process, usually caused by a bad API call such as entering
  // or leaving a destroyed critical section. If the last error is 
  // ERROR_NOT_OWNER, then that means that EndSynchronized() was called
  // without calling BeginSynchronized() first. IN THIS CASE, THE 
  // INTERNAL STATE OF THE OBJECT IS CORRUPTED. THERFORE, THIS MUST 
  // ALSO BE CONSIDERED A FATAL ERROR.
  BOOL EndSynchronized();

  // Waits for a notification. The first argument is the timeout
  // in milliseconds. The alertable flag must be true if the wait
  // should complete when a user APC is queued.
  //
  // In case of success, the return value is WAIT_OBJECT_0. See
  // the documentation of the Win32 function SignalObjectAndWait()
  // for an explanation of the other possible return values.
  //
  // NOTE: The return value WAIT_FAILED indicates an irrecoverably bad
  // state of the process, usually caused by a bad API call such as 
  // entering or leaving a destroyed critical section. If the last 
  // error is ERROR_NOT_OWNER, then that means that Wait() was called 
  // without acquiring the lock first. IN THIS CASE, THE INTERNAL STATE
  // OF THE OBJECT IS CORRUPTED. THERFORE, THIS MUST ALSO BE CONSIDERED
  // A FATAL ERROR.  
  DWORD Wait(
    DWORD dwMillisecondsTimeout,
    BOOL bAlertable
    );

  // Notifies one or all of the waiting threads, if any, according
  // to the notification type specified during creation. In case of 
  // success, the return value is TRUE. In case of failure, call 
  // GetLastError() to obtain error information.
  //
  // NOTE: The return value FALSE indicates an irrecoverrably bad state
  // of the process, usually caused by a bad API call such entering or
  // leaving a destroyed critical section.
  BOOL Pulse();

  // Checks if the handles to the synch objects are non-NULL.
  bool IsValid();

private:

  // Private helper functions
  // ========================
  
  // Closes handles.
  void Cleanup();

  // Data members
  // ============

  // Handle to notification event.
  HANDLE m_hNotificationEvent;

  // Handle to synchronization mutex. NOTE: This cannot
  // be replaced with a critical section, because a critical
  // section cannot be used in SignalObjectAndWait().
  HANDLE m_hSynchronizationMutex;

  // The monitor must keep track of how many times the lock
  // has been acquired, because Win32 does not divulge this
  // information to the client programmer.
  int m_nLockCount;

};

#endif
//
// End multiple inclusion protection
