/**********************************************************************
*
COPYRIGHT (C) 2000 Thomas Becker

PERMISSION NOTICE:

PERMISSION TO USE, COPY, MODIFY, AND DISTRIBUTE THIS SOFTWARE FOR ANY
PURPOSE AND WITHOUT FEE IS HEREBY GRANTED, PROVIDED THAT ALL COPIES
ARE ACCOMPANIED BY THE COMPLETE MACHINE-READABLE SOURCE CODE, ALL
MODIFIED FILES CARRY PROMINENT NOTICES AS TO WHEN AND BY WHOM THEY
WERE MODIFIED, THE ABOVE COPYRIGHT NOTICE, THIS PERMISSION NOTICE AND
THE NO-WARRANTY NOTICE BELOW APPEAR IN ALL SOURCE FILES AND IN ALL
SUPPORTING DOCUMENTATION.

NO-WARRANTY NOTICE:

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
MERCHANTABILITY ARE HEREBY DISCLAIMED.
*
**********************************************************************/
 
//
// File: MultiSetMonitor.cpp
//
// Author: Thomas Becker
//
// Created: Dec 2000
//
// Description: Implementation of class MultiSetMonitor
//

#include"stdafx.h"
#include "MultiSetMonitor.h"

///////////////////////////////////////////////////////////////////////
//
MultiSetMonitor::MultiSetMonitor() : 
  m_hSynchronizationMutex(NULL),
  m_nLockCount(0)
  {/* All initialization takes place in the Create() method. */}

///////////////////////////////////////////////////////////////////////
//
MultiSetMonitor::~MultiSetMonitor()
{ Cleanup(); }

///////////////////////////////////////////////////////////////////////
//
bool MultiSetMonitor::Create(
  int nNumNotifySets,
  int nNumNotifyAllSets
  )
{
  // Check if already created.
  if( IsValid() ) 
  {
    ::SetLastError(ERROR_ALREADY_EXISTS);
    return false;
  }
  
  // Create the notify events.
  //
  m_vectNotifyEvents.reserve(nNumNotifySets);
  for( int i = 0; i < nNumNotifySets; ++i )
  {
    HANDLE hNotificationEvent = ::CreateEvent(
      NULL, // no security
      FALSE, // auto-reset event
      FALSE, // initially unsignalled
      NULL // string name
      );
    //
    if( NULL == hNotificationEvent ) 
    {
      // Store last error as it may be overwritten by Cleanup().
      DWORD dwLastError = GetLastError();
      Cleanup();
      // Restore last error.
      SetLastError(dwLastError);
      return false;
    }
    m_vectNotifyEvents.push_back(hNotificationEvent);
  }

  // Create the notify all events.
  //
  m_vectNotifyAllEvents.reserve(nNumNotifyAllSets);
  for( i = 0; i < nNumNotifyAllSets; ++i )
  {
    HANDLE hNotificationEvent = ::CreateEvent(
      NULL, // no security
      TRUE, // manual reset event
      FALSE, // initially unsignalled
      NULL // string name
      );
    //
    if( NULL == hNotificationEvent ) 
    {
      // Store last error as it may be overwritten by Cleanup().
      DWORD dwLastError = GetLastError();
      Cleanup();
      // Restore last error.
      SetLastError(dwLastError);
      return false;
    }
    m_vectNotifyAllEvents.push_back(hNotificationEvent);
  }

  // Create the synchronization mutex.
  m_hSynchronizationMutex = ::CreateMutex(
    NULL, // no security
    FALSE, // initially not owned
    NULL // string name
    );
  //
  if( NULL == m_hSynchronizationMutex ) 
  {
    // Store last error as it may be overwritten by Cleanup().
    DWORD dwLastError = GetLastError();
    Cleanup();
    // Restore last error.
    SetLastError(dwLastError);
    return false;
  }

  return true;
}

///////////////////////////////////////////////////////////////////////
//
DWORD MultiSetMonitor::BeginSynchronized()
{
  // Wait for mutex.
  DWORD dwWaitRetVal = ::WaitForSingleObject(
    m_hSynchronizationMutex, 
    INFINITE
    );

  // If the wait completed successfully, record the lock acquisition 
  // for proper release in Wait().
  if( WAIT_OBJECT_0 == dwWaitRetVal )
    ++m_nLockCount;

  return dwWaitRetVal;
}

///////////////////////////////////////////////////////////////////////
//
BOOL MultiSetMonitor::EndSynchronized()
{
  // Record the lock release for proper release in Wait().
  --m_nLockCount;

  // Release mutex.
  BOOL bReleaseRetVal = ::ReleaseMutex(m_hSynchronizationMutex);

  return bReleaseRetVal;
}

///////////////////////////////////////////////////////////////////////
//
DWORD MultiSetMonitor::WaitInNotifySet(
  int nWaitSet,
  DWORD dwMillisecondsTimeout,
  BOOL bAlertable
  )
{
  // Check for valid set index.
  if( 0 > nWaitSet || m_vectNotifyEvents.size() <= nWaitSet )
  {
    ::SetLastError(ERROR_INVALID_INDEX);
    return WAIT_FAILED;
  }

  // Store the current lock count for re-acquisition.
  int nThisThreadsLockCount = m_nLockCount;
  m_nLockCount = 0;

  // Release the mutex enough times so that it is owned exactly once.
  //
  for( int i=0; i<nThisThreadsLockCount-1; ++i )
  {
    BOOL bReleaseRetVal = ::ReleaseMutex(m_hSynchronizationMutex);
    // If the release failed, we must return WAIT_FAILED. The client
    // will call GetLastError() to find out what really happened.
    if( ! bReleaseRetVal )
      return WAIT_FAILED;
  }
  
  // Atomically release lock and wait for event to be pulsed.
  DWORD dwWaitResult = ::SignalObjectAndWait(
    m_hSynchronizationMutex,
    m_vectNotifyEvents[nWaitSet],
    dwMillisecondsTimeout,
    bAlertable
    );

  // If the wait failed, store last error. And YES, the documentation
  // specifies 0xFFFFFFFF, not WAIT_FAILED as the return value on 
  // failure.
  DWORD dwWaitError;
  if( 0xFFFFFFFF == dwWaitResult )
  {
    dwWaitResult = WAIT_FAILED; // That's what it should have been.
    dwWaitError = ::GetLastError();
  }

  // Re-acquire the lock. NOTE: Do not use an alertable wait
  // here! The lock must be reacquired when the wait is 
  // interrupted
  //
  DWORD dwLockResult;
  for( int j=0; j<nThisThreadsLockCount; ++j )
  {
    dwLockResult = ::WaitForSingleObject(
      m_hSynchronizationMutex,
      INFINITE
      );
    if( WAIT_OBJECT_0 != dwLockResult )
      break;
  }

  // If either one of the wait operations failed, we're irrecoverably
  // screwed.
  if( WAIT_FAILED == dwWaitResult || 
      WAIT_FAILED == dwLockResult ||
      WAIT_ABANDONED == dwLockResult )
  {
    // If the second wait was successful, we must restore the error
    // code for the client.
    if( WAIT_OBJECT_0 == dwLockResult )
    {
      m_nLockCount = nThisThreadsLockCount;
      ::SetLastError(dwWaitError);
      return dwWaitResult;
    }

    return dwLockResult;
  }

  // What matters to the caller is the return value of the wait.
  m_nLockCount = nThisThreadsLockCount;
  return dwWaitResult;

}

///////////////////////////////////////////////////////////////////////
//
DWORD MultiSetMonitor::WaitInNotifyAllSet(
  int nWaitSet,
  DWORD dwMillisecondsTimeout,
  BOOL bAlertable
  )
{
  // Check for valid set index.
  if( 0> nWaitSet || m_vectNotifyAllEvents.size() <= nWaitSet )
  {
    ::SetLastError(ERROR_INVALID_INDEX);
    return WAIT_FAILED;
  }

  // Store the current lock count for re-acquisition.
  int nThisThreadsLockCount = m_nLockCount;
  m_nLockCount = 0;

  // Release the mutex enough times so that it is owned exactly once.
  //
  for( int i=0; i<nThisThreadsLockCount-1; ++i )
  {
    BOOL bReleaseRetVal = ::ReleaseMutex(m_hSynchronizationMutex);
    // If the release failed, we must return WAIT_FAILED. The client
    // will call GetLastError() to find out what really happened.
    if( ! bReleaseRetVal )
      return WAIT_FAILED;
  }
  
  // Atomically release lock and wait for event to be pulsed.
  DWORD dwWaitResult = ::SignalObjectAndWait(
    m_hSynchronizationMutex,
    m_vectNotifyAllEvents[nWaitSet],
    dwMillisecondsTimeout,
    bAlertable
    );

  // If the wait failed, store last error. And YES, the documentation
  // specifies 0xFFFFFFFF, not WAIT_FAILED as the return value on 
  // failure.
  DWORD dwWaitError;
  if( 0xFFFFFFFF == dwWaitResult )
  {
    dwWaitResult = WAIT_FAILED; // That's what it should have been.
    dwWaitError = ::GetLastError();
  }

  // Re-acquire the lock. NOTE: Do not use an alertable wait
  // here! The lock must be reacquired when the wait is 
  // interrupted
  //
  DWORD dwLockResult;
  for( int j=0; j<nThisThreadsLockCount; ++j )
  {
    dwLockResult = ::WaitForSingleObject(
      m_hSynchronizationMutex,
      INFINITE
      );
    if( WAIT_OBJECT_0 != dwLockResult )
      break;
  }

  // If either one of the wait operations failed, we're irrecoverably
  // screwed.
  if( WAIT_FAILED == dwWaitResult || 
      WAIT_FAILED == dwLockResult ||
      WAIT_ABANDONED == dwLockResult )
  {
    // If the second wait was successful, we must restore the error
    // code for the client.
    if( WAIT_OBJECT_0 == dwLockResult )
    {
      m_nLockCount = nThisThreadsLockCount;
      ::SetLastError(dwWaitError);
      return dwWaitResult;
    }

    return dwLockResult;
  }

  // What matters to the caller is the return value of the wait.
  m_nLockCount = nThisThreadsLockCount;
  return dwWaitResult;

}


///////////////////////////////////////////////////////////////////////
//
BOOL MultiSetMonitor::Notify(
  int nWaitSet)
{
  // Check for valid set index.
  if( 0> nWaitSet || m_vectNotifyEvents.size() <= nWaitSet )
  {
    ::SetLastError(ERROR_INVALID_INDEX);
    return FALSE;
  }

  return ::PulseEvent(m_vectNotifyEvents[nWaitSet]);
}

///////////////////////////////////////////////////////////////////////
//
BOOL MultiSetMonitor::NotifyAll(
  int nWaitSet)
{
  // Check for valid set index.
  if( 0> nWaitSet || m_vectNotifyAllEvents.size() <= nWaitSet )
  {
    ::SetLastError(ERROR_INVALID_INDEX);
    return FALSE;
  }

  return ::PulseEvent(m_vectNotifyAllEvents[nWaitSet]);
}

///////////////////////////////////////////////////////////////////////
//
void MultiSetMonitor::Cleanup()
{
  
  //
  // NOTE: This function is called only in the destructor and when
  // some Win32 call has failed. Therefore, we do not check the
  // return value of CloseHandle() here.
  //

  // Close the notify handles and free the memory.
  //
  std::vector<HANDLE>::const_iterator it_run = 
    m_vectNotifyEvents.begin();
  std::vector<HANDLE>::const_iterator it_end = 
    m_vectNotifyEvents.end();
  //
  for( ; it_run < it_end; ++it_run )
    CloseHandle(*it_run);
  //
  m_vectNotifyEvents.swap(std::vector<HANDLE>());

  // Close the notify all handles and free the memory.
  //
  it_run = m_vectNotifyAllEvents.begin();
  it_end = m_vectNotifyAllEvents.end();
  //
  for( ; it_run < it_end; ++it_run )
    CloseHandle(*it_run);
  //
  m_vectNotifyAllEvents.swap(std::vector<HANDLE>());

  if(NULL != m_hSynchronizationMutex) {
    CloseHandle(m_hSynchronizationMutex);
    m_hSynchronizationMutex = NULL;
  }
}

///////////////////////////////////////////////////////////////////////
//
bool MultiSetMonitor::IsValid()
{
  return NULL != m_hSynchronizationMutex;
}
