/**********************************************************************
*
COPYRIGHT (C) 2000 Thomas Becker

PERMISSION NOTICE:

PERMISSION TO USE, COPY, MODIFY, AND DISTRIBUTE THIS SOFTWARE FOR ANY
PURPOSE AND WITHOUT FEE IS HEREBY GRANTED, PROVIDED THAT ALL COPIES
ARE ACCOMPANIED BY THE COMPLETE MACHINE-READABLE SOURCE CODE, ALL
MODIFIED FILES CARRY PROMINENT NOTICES AS TO WHEN AND BY WHOM THEY
WERE MODIFIED, THE ABOVE COPYRIGHT NOTICE, THIS PERMISSION NOTICE AND
THE NO-WARRANTY NOTICE BELOW APPEAR IN ALL SOURCE FILES AND IN ALL
SUPPORTING DOCUMENTATION.

NO-WARRANTY NOTICE:

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
MERCHANTABILITY ARE HEREBY DISCLAIMED.
*
**********************************************************************/
 
//
// File: MultiSetMonitor.h
//
// Author: Thomas Becker
//
// Created: Dec 2000
//
// Description: Declaration of class MultiSetMonitor
//

// Multiple inclusion protection
//
#ifndef __MULTI_SET_MONITOR_INCLUDED__
#define __MULTI_SET_MONITOR_INCLUDED__

#include"stdafx.h"
#include<vector>

///////////////////////////////////////////////////////////////////////
//
// The class MultiSetMonitor implements a Java-style monitor in Win32. 
// This monitor goes beyond the functionality found in Java monitors 
// insofar as it supports multiple wait set. On the other hand, each
// wait set only supports Notify() or NotifyAll(), but not both.
//
// This class internally uses the Win32 function SignalObjectAndWait(),
// which is supported only under Windows NT 4.0 and higher. Therefore,
// you must define _WIN32_WINNT 0x0400 for this to work.
//
class MultiSetMonitor
{
public:

  // Public Construction and Destruction
  // ===================================
  
  // Default constructor. The MultiSetMonitor is not valid until the
  // Create() member function has been called.
  MultiSetMonitor();

  // The destructor closes all handles.
  //
  // NOTE: This class is not meant to be derived from. If you must 
  // derive from this class, you should make the destructor virtual.
  ~MultiSetMonitor();

  // Creates the all synchronization objects. The first argument is
  // the number of wait sets that can respond to Notify(). The second
  // argument is the number of wait sets that can respond to 
  // NotifyAll().
  //
  // Returns true in case of success, false otherwise. In case of 
  // failure, call GetLastError() to obtain error information.
  bool Create(
    int nNumNotifySets,
    int nNumNotifyAllSets
    );

  // Operations
  // ==========
  
  // Acquires the lock once. In case of sucess, the return value is
  // WAIT_OBJECT_0. See the documentation of the Win32 function
  // WaitForSingleObject() for an explanation of the other possible
  // return values. The object being waited on is a mutex. 
  //
  // NOTE: The return value WAIT_FAILED indicates an irrecoverably
  // bad state of the process, usually caused by a bad API call such as
  // entering or leaving a destroyed critical section.
  DWORD BeginSynchronized();

  // Releases the lock once. In case of success, the return value is 
  // TRUE. In case of failure, call GetLastError() to obtain error
  // information.
  //
  // NOTE: The return value FALSE indicates an irrecoverably bad state
  // of the process, usually caused by a bad API call such as entering
  // or leaving a destroyed critical section. If the last error is 
  // ERROR_NOT_OWNER, then that means that EndSynchronized() was called
  // without calling BeginSynchronized() first. IN THIS CASE, THE 
  // INTERNAL STATE OF THE OBJECT IS CORRUPTED. THERFORE, THIS MUST 
  // ALSO BE CONSIDERED A FATAL ERROR.
  BOOL EndSynchronized();

  // Waits for a notification in one of the wait sets that can respond
  // to Notify(). The first argument specifies the index of the wait 
  // set. This index must satisfy
  //
  // 0 <= index < nNumNotifySets
  //
  // where nNumNotifySets is the number of wait sets that can respond
  // to Notify() as specified in the call to Create().
  //
  // The second argument is the timeout in milliseconds. The alertable
  // flag must be true if the wait should complete when a user APC is 
  // queued.
  //
  // In case of success, the return value is WAIT_OBJECT_0. See
  // the documentation of the Win32 function SignalObjectAndWait()
  // for an explanation of the other possible return values.
  //
  // NOTE: In most cases, the return value WAIT_FAILED indicates an 
  // irrecoverably bad state of the process, usually caused by a bad 
  // API call such as entering or leaving a destroyed critical section.
  // If the last error is ERROR_NOT_OWNER, then that means that Wait() 
  // was called without acquiring the lock first. IN THIS CASE, THE 
  // INTERNAL STATE OF THE OBJECT IS CORRUPTED. THERFORE, THIS MUST 
  // ALSO BE CONSIDERED A FATAL ERROR. If the last error is 
  // ERROR_INVALID_INDEX, then the index of the wait set is not within
  // the range of the number of wait sets as specified in Create().
  // In this case, the state of the object remains intact.
  DWORD WaitInNotifySet(
    int nNotifyWaitSet,
    DWORD dwMillisecondsTimeout,
    BOOL bAlertable
    );

  // Waits for a notification in one of the wait sets that can respond
  // to NotifyAll(). The first argument specifies the index of the wait 
  // set. This index must satisfy
  //
  // 0 <= index < nNumNotifyAllSets
  //
  // where nNumNotifyAllSets is the number of wait sets that can 
  // respond to NotifyAll() as specified in the call to Create().
  //
  // The second argument is the timeout in milliseconds. The alertable
  // flag must be true if the wait should complete when a user APC is 
  // queued.
  //
  // In case of success, the return value is WAIT_OBJECT_0. See
  // the documentation of the Win32 function SignalObjectAndWait()
  // for an explanation of the other possible return values.
  //
  // NOTE: In most cases, the return value WAIT_FAILED indicates an 
  // irrecoverably bad state of the process, usually caused by a bad 
  // API call such as entering or leaving a destroyed critical section.
  // If the last error is ERROR_NOT_OWNER, then that means that Wait() 
  // was called without acquiring the lock first. IN THIS CASE, THE 
  // INTERNAL STATE OF THE OBJECT IS CORRUPTED. THERFORE, THIS MUST 
  // ALSO BE CONSIDERED A FATAL ERROR. If the last error is 
  // ERROR_INVALID_INDEX, then the index of the wait set is not within
  // the range of the number of wait sets as specified in Create().
  // In this case, the state of the object remains intact.
  DWORD WaitInNotifyAllSet(
    int nNotifyAllWaitSet,
    DWORD dwMillisecondsTimeout,
    BOOL bAlertable
    );

  // Notifies one of the threads, if any, that are currently waiting in
  // the set whose index is specified in the argument. This index must
  // satisfy
  //
  // 0 <= index < nNumNotifySets
  //  
  // where nNumNotifySets is the number of wait sets that can 
  // respond to Notify() as specified in the call to Create().
  //
  // In case of success,  the return value is TRUE. In case of failure,
  // call GetLastError() to obtain error information.
  //
  // NOTE: In most cases, the return value FALSE indicates an 
  // irrecoverably bad state of the process, usually caused by a bad 
  // API call such as entering or leaving a destroyed critical section.
  // There is only one exception to this rule: if the last error is 
  // ERROR_INVALID_INDEX, then the index of the wait set is not within
  // the range of the number of wait sets as specified in Create().
  BOOL Notify(int nWaitSet);

  // Notifies all of the threads, if any, that are currently waiting in
  // the set whose index is specified in the argument. This index must
  // satisfy
  //
  // 0 <= index < nNumNotifyAllSets
  //  
  // where nNumNotifyAllSets is the number of wait sets that can 
  // respond to NotifyAll() as specified in the call to Create().
  //
  // In case of success,  the return value is TRUE. In case of failure,
  // call GetLastError() to obtain error information.
  //
  // NOTE: In most cases, the return value FALSE indicates an 
  // irrecoverably bad state of the process, usually caused by a bad 
  // API call such as entering or leaving a destroyed critical section.
  // There is only one exception to this rule: if the last error is 
  // ERROR_INVALID_INDEX, then the index of the wait set is not within
  // the range of the number of wait sets as specified in Create().
  BOOL NotifyAll(int nWaitSet);

  // Checks if the handles to the synch objects are non-NULL.
  bool IsValid();

private:

  // Private helper functions
  // ========================
  
  // Closes handles and frees memory.
  //
  // NOTE: This function is called only in the destructor and when
  // some Win32 call has failed. Therefore, we do not check the
  // return value of CloseHandle() here.
  void Cleanup();

  // Data members
  // ============

  // Vector of handles to notification events, one for
  // each notify set.
  std::vector<HANDLE> m_vectNotifyEvents;

  // Vector of handles to notification events, one for
  // each notify set.
  std::vector<HANDLE> m_vectNotifyAllEvents;

  // Handle to synchronization mutex. NOTE: This cannot
  // be replaced with a critical section, because critical
  // sections cannot be used with SignalObjectAndWait().
  HANDLE m_hSynchronizationMutex;

  // The monitor must keep track of how many times the lock
  // has been acquired, because Win32 does not divulge this
  // information to the client programmer.
  int m_nLockCount;

};

#endif
//
// End multiple inclusion protection
