/**********************************************************************
*
COPYRIGHT (C) 2000 Thomas Becker

PERMISSION NOTICE:

PERMISSION TO USE, COPY, MODIFY, AND DISTRIBUTE THIS SOFTWARE FOR ANY
PURPOSE AND WITHOUT FEE IS HEREBY GRANTED, PROVIDED THAT ALL COPIES
ARE ACCOMPANIED BY THE COMPLETE MACHINE-READABLE SOURCE CODE, ALL
MODIFIED FILES CARRY PROMINENT NOTICES AS TO WHEN AND BY WHOM THEY
WERE MODIFIED, THE ABOVE COPYRIGHT NOTICE, THIS PERMISSION NOTICE AND
THE NO-WARRANTY NOTICE BELOW APPEAR IN ALL SOURCE FILES AND IN ALL
SUPPORTING DOCUMENTATION.

NO-WARRANTY NOTICE:

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
MERCHANTABILITY ARE HEREBY DISCLAIMED.
*
**********************************************************************/
 
//
// File: NotifyMonitor.h
//
// Author: Thomas Becker
//
// Created: Dec 2000
//
// Description: Declaration of class NotifyMonitor
//

// Multiple inclusion protection
//
#ifndef __NOTIFY_MONITOR_INCLUDED__
#define __NOTIFY_MONITOR_INCLUDED__

#include"stdafx.h"
#include"MonitorImpl.h"

///////////////////////////////////////////////////////////////////////
//
// The class NotifyMonitor implements a Java-style monitor with the 
// restriction that only single notifications are possible, i.e., there
// is no NotifyAll(). Technically, the class uses an implementation
// member object of type MonitorImpl. That way, NotifyMonitor can
// share the implementation with its cousin, class NotifyAllMonitor.
//
// This class internally uses the Win32 function SignalObjectAndWait(),
// which is supported only under Windows NT 4.0 and higher. Therefore,
// you must define _WIN32_WINNT 0x0400 for this to work.
//
class NotifyMonitor
{
public:

  // Public Construction and Destruction
  // ===================================
  
  // Uses the compiler-generated default constructor. The monitor is 
  // not valid until the Create() member function has been called.

  // Uses the compiler-generated default destructor. The destructor of 
  // the implementation object handles cleanup. 
  //
  // NOTE: This class is not meant to be derived from. If you must 
  // derive from this class, you should uncomment the line below.
  //
  // virtual ~NotifyMonitor() {}

  // Creates the implementation object. Returns true in case of
  // success, false otherwise. In case of failure, call GetLastError()
  // to obtain error information.
  bool Create()
  { return m_Impl.Create(MonitorImpl::enNotify); }

  // Acquires the lock once. In case of sucess, the return value is
  // WAIT_OBJECT_0. See the documentation of the Win32 function
  // WaitForSingleObject() for an explanation of the other possible
  // return values. The object being waited on is a mutex. 
  //
  // NOTE: The return value WAIT_FAILED indicates an irrecoverably
  // bad state of the process, usually caused by a bad API call such as
  // entering or leaving a destroyed critical section.
  DWORD BeginSynchronized()
  { return m_Impl.BeginSynchronized(); }

  // Releases the lock once. In case of success, the return value is 
  // TRUE. In case of failure, call GetLastError() to obtain error
  // information.
  //
  // NOTE: The return value FALSE indicates an irrecoverably bad state
  // of the process, usually caused by a bad API call such as entering
  // or leaving a destroyed critical section. If the last error is 
  // ERROR_NOT_OWNER, then that means that EndSynchronized() was called
  // without calling BeginSynchronized() first. IN THIS CASE, THE 
  // INTERNAL STATE OF THE OBJECT IS CORRUPTED. THERFORE, THIS MUST 
  // ALSO BE CONSIDERED A FATAL ERROR.
  BOOL EndSynchronized()
  { return m_Impl.EndSynchronized(); }

  // Waits for a notification. The first argument is the timeout
  // in milliseconds. The alertable flag must be true if the wait
  // should complete when a user APC is queued.
  //
  // In case of success, the return value is WAIT_OBJECT_0. See
  // the documentation of the Win32 function SignalObjectAndWait()
  // for an explanation of the other possible return values.
  //
  // NOTE: The return value WAIT_FAILED indicates an irrecoverably bad
  // state of the process, usually caused by a bad API call such as 
  // entering or leaving a destroyed critical section. If the last 
  // error is ERROR_NOT_OWNER, then that means that Wait() was called 
  // without acquiring the lock first. IN THIS CASE, THE INTERNAL STATE
  // OF THE OBJECT IS CORRUPTED. THERFORE, THIS MUST ALSO BE CONSIDERED
  // A FATAL ERROR.  
  DWORD Wait(
    DWORD dwMillisecondsTimeout = INFINITE,
    BOOL bAlertable = FALSE
    )
  { return m_Impl.Wait(dwMillisecondsTimeout, bAlertable); }

  // Notifies one of the waiting threads, if any. In case of success, 
  // the return value is TRUE. In case of failure, call GetLastError()
  // to obtain error information.
  //
  // NOTE: The return value FALSE indicates an irrecoverrably bad state
  // of the process, usually caused by a bad API call such entering or
  // leaving a destroyed critical section.
  BOOL Notify()
  { return m_Impl.Pulse(); }

private:

  // Data members
  // ============

  // Implementation object
  MonitorImpl m_Impl;

};

#endif
//
// End multiple inclusion protection
