/**********************************************************************
*
COPYRIGHT (C) 2000 Thomas Becker

PERMISSION NOTICE:

PERMISSION TO USE, COPY, MODIFY, AND DISTRIBUTE THIS SOFTWARE FOR ANY
PURPOSE AND WITHOUT FEE IS HEREBY GRANTED, PROVIDED THAT ALL COPIES
ARE ACCOMPANIED BY THE COMPLETE MACHINE-READABLE SOURCE CODE, ALL
MODIFIED FILES CARRY PROMINENT NOTICES AS TO WHEN AND BY WHOM THEY
WERE MODIFIED, THE ABOVE COPYRIGHT NOTICE, THIS PERMISSION NOTICE AND
THE NO-WARRANTY NOTICE BELOW APPEAR IN ALL SOURCE FILES AND IN ALL
SUPPORTING DOCUMENTATION.

NO-WARRANTY NOTICE:

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
MERCHANTABILITY ARE HEREBY DISCLAIMED.
*
**********************************************************************/
 
//
// File: ReadWriteLock.cpp
//
// Author: Thomas Becker
//
// Created: Dec 2000
//
// Description: Implementation of class ReadWriteLock
//

#include"stdafx.h"
#include "ReadWriteLock.h"
#include "time.h"

///////////////////////////////////////////////////////////////////////
//
ReadWriteLock::ReadWriteLock() : 
  m_nNumActiveReaders(0),
  m_nNumWaitingWriters(0),
  m_bWriteInProgress(false)
  {/* All initialization takes place in the Create() method. */}

///////////////////////////////////////////////////////////////////////
//
bool ReadWriteLock::Create()
{
  // Check if already created.
  if( IsValid() ) {
    ::SetLastError(ERROR_ALREADY_EXISTS);
    return false;
  }
  
  // Create the monitor.
  //
  if( ! m_monReadWrite.Create(
          1, // one notify set
          1 // one notify all set
        ) ) 
  {
    return false;
  }

  return true;
}

///////////////////////////////////////////////////////////////////////
//
DWORD ReadWriteLock::EnterRead(
  DWORD dwMillisecondsTimeout/*=INFINITE*/,
  BOOL bAlertable/*=FALSE*/
  )
{

  // Acquire monitor lock.
  //
  DWORD dwBeginSyncError = m_monReadWrite.BeginSynchronized();
  if( WAIT_OBJECT_0 != dwBeginSyncError )
    return dwBeginSyncError;
  
  // Wait for the number of waiting or active writers to reach
  // zero, then increment the number of active readers.
  //
  DWORD dwWaitRetVal = WAIT_OBJECT_0;
  DWORD dwWaitError;
  while( 0 < m_nNumWaitingWriters || m_bWriteInProgress )
  {
    dwWaitRetVal = m_monReadWrite.WaitInNotifyAllSet(
      0,
      dwMillisecondsTimeout, 
      bAlertable
      );
    
    // If the wait failed, we're through.
    if( WAIT_FAILED == dwWaitRetVal )
    {
      dwWaitError = ::GetLastError();
      m_monReadWrite.EndSynchronized();
      ::SetLastError(dwWaitError);
      return dwWaitRetVal;
    }
  }
  ++m_nNumActiveReaders;

  // Release monitor lock.
  if( ! m_monReadWrite.EndSynchronized() )
    return WAIT_FAILED;
 
  return dwWaitRetVal;
}

///////////////////////////////////////////////////////////////////////
//
BOOL ReadWriteLock::LeaveRead()
{
  // Acquire monitor lock.
  //
  DWORD dwBeginSyncError = m_monReadWrite.BeginSynchronized();
  if( WAIT_OBJECT_0 != dwBeginSyncError )
    return FALSE;
  
  // Decrement the number of active readers. If it is now zero,
  // notify on the writer set.
  //
  BOOL bNotifyResult = TRUE;
  DWORD dwNotifyError;
  if( 0 == --m_nNumActiveReaders )
  {
    // Notify on the writer set.
    bNotifyResult = m_monReadWrite.Notify(0);
    // If the notification failed, store the last error 
    // for later restoration.
    if( ! bNotifyResult )
      dwNotifyError = ::GetLastError();
  }

  // Release monitor lock.
  if( ! m_monReadWrite.EndSynchronized() )
    return FALSE;
  
  // If the notification failed, restore the last error for the client.
  if( ! bNotifyResult )
    ::SetLastError(dwNotifyError);
  return bNotifyResult;
}

///////////////////////////////////////////////////////////////////////
//
DWORD ReadWriteLock::EnterWrite(
  DWORD dwMillisecondsTimeout/*=INFINITE*/,
  BOOL bAlertable/*=FALSE*/
  )
{

  // Acquire monitor lock.
  //
  DWORD dwBeginSyncError = m_monReadWrite.BeginSynchronized();
  if( WAIT_OBJECT_0 != dwBeginSyncError )
    return dwBeginSyncError;
  
  // Increment the number of waiting writers, then
  // wait for the number of readers to go to zero.
  // Upon completing the wait, decrement the number
  // of waiting writers and set the write-in-progress
  // flag to true.
  //
  DWORD dwWaitRetVal = WAIT_OBJECT_0;
  DWORD dwWaitError;
  ++m_nNumWaitingWriters;
  while( 0 < m_nNumActiveReaders || m_bWriteInProgress )
  {
    dwWaitRetVal = m_monReadWrite.WaitInNotifySet(
      0,
      dwMillisecondsTimeout, 
      bAlertable
      );
    
    // If the wait failed, we're through.
    if( WAIT_FAILED == dwWaitRetVal )
    {
      dwWaitError = ::GetLastError();
      m_monReadWrite.EndSynchronized();
      ::SetLastError(dwWaitError);
      return dwWaitRetVal;
    }
  }
  --m_nNumWaitingWriters;
  m_bWriteInProgress = true;

  // Release monitor lock.
  if( ! m_monReadWrite.EndSynchronized() )
    return WAIT_FAILED;

  return dwWaitRetVal;

}

///////////////////////////////////////////////////////////////////////
//
BOOL ReadWriteLock::LeaveWrite()
{
  
  // Acquire monitor lock.
  //
  DWORD dwBeginSyncError = m_monReadWrite.BeginSynchronized();
  if( WAIT_OBJECT_0 != dwBeginSyncError )
    return FALSE;
  
  // Set the write-in-progress flag to false. If there are
  // no waiting writers, notify all on the reader set. Else,
  // notify on the writer set.
  //
  BOOL bNotifyResult = TRUE;
  DWORD dwNotifyError;
  m_bWriteInProgress = false;
  if( 0 == m_nNumWaitingWriters )
  {
    bNotifyResult = m_monReadWrite.NotifyAll(0);
    // If the notification failed, store the last error 
    // for later restoration.
    if( ! bNotifyResult )
      dwNotifyError = ::GetLastError();
  }
  else
  {
    bNotifyResult = m_monReadWrite.Notify(0);
    // If the notification failed, store the last error 
    // for later restoration.
    if( ! bNotifyResult )
      dwNotifyError = ::GetLastError();
  }

  // Release monitor lock.
  if( ! m_monReadWrite.EndSynchronized() )
    return FALSE;

  // If the notification failed, restore the last error for the client.
  if( ! bNotifyResult )
    ::SetLastError(dwNotifyError);
  return bNotifyResult;
}

///////////////////////////////////////////////////////////////////////
//
bool ReadWriteLock::IsValid()
{
  return m_monReadWrite.IsValid();
}

