/**********************************************************************
*
COPYRIGHT (C) 2000 Thomas Becker

PERMISSION NOTICE:

PERMISSION TO USE, COPY, MODIFY, AND DISTRIBUTE THIS SOFTWARE FOR ANY
PURPOSE AND WITHOUT FEE IS HEREBY GRANTED, PROVIDED THAT ALL COPIES
ARE ACCOMPANIED BY THE COMPLETE MACHINE-READABLE SOURCE CODE, ALL
MODIFIED FILES CARRY PROMINENT NOTICES AS TO WHEN AND BY WHOM THEY
WERE MODIFIED, THE ABOVE COPYRIGHT NOTICE, THIS PERMISSION NOTICE AND
THE NO-WARRANTY NOTICE BELOW APPEAR IN ALL SOURCE FILES AND IN ALL
SUPPORTING DOCUMENTATION.

NO-WARRANTY NOTICE:

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
MERCHANTABILITY ARE HEREBY DISCLAIMED.
*
**********************************************************************/
 
//
// File: ReadWriteLock.h
//
// Author: Thomas Becker
//
// Created: Dec 2000
//
// Description: Declaration of class ReadWriteLock
//

// Multiple inclusion protection
//
#ifndef __READ_WRITE_LOCK_INCLUDED__
#define __READ_WRITE_LOCK_INCLUDED__

#include"stdafx.h"
#include"MultiSetMonitor.h"

///////////////////////////////////////////////////////////////////////
//
// The class ReadWriteLock implements a read-write lock. A read-write
// lock consists of two locks, the read lock and the write lock. The
// two are connected as follows:
//
// No thread can obtain the read lock when some thread owns or is 
// waiting on the write lock.
//
// No thread can obtain the write lock when some thread owns the read 
// or write lock.
//
// Note that under this policy, writers can starve all reading.
//
// This class internally uses the Win32 function SignalObjectAndWait(),
// which is supported only under Windows NT 4.0 and higher. Therefore,
// you must define _WIN32_WINNT 0x0400 for this to work.
//
class ReadWriteLock
{
public:

  // Public Construction and Destruction
  // ===================================
  
  // Default constructor. The ReadWriteLock is not valid until the
  // Create() member function has been called.
  ReadWriteLock();

  // Uses the compiler-generated default destructor. The destructor of 
  // the implementation object handles cleanup. 
  //
  // NOTE: This class is not meant to be derived from. If you must 
  // derive from this class, you should uncomment the line below.
  //
  // virtual ~ReadWriteLock() {}

  // Creates the all synchronization objects.
  //
  // Returns true in case of success, false otherwise. In case of 
  // failure, call GetLastError() to obtain error information.
  bool Create();

  // Operations
  // ==========
  
  // Acquires the read lock. The first argument is the timeout in 
  // milliseconds. The alertable flag must be true if the wait should 
  // complete when a user APC is queued.
  //
  // In case of success, the return value is WAIT_OBJECT_0. See
  // the documentation of the Win32 function SignalObjectAndWait()
  // for an explanation of the other possible return values.
  //
  // NOTE: The return value WAIT_FAILED indicates an irrecoverably bad
  // state of the process, usually caused by a bad API call such as 
  // entering or leaving a destroyed critical section.
  DWORD EnterRead(
    DWORD dwMillisecondsTimeout = INFINITE,
    BOOL bAlertable = FALSE
    );

  // Releases the read lock. In case of success,  the return value is 
  // TRUE. In case of failure, call GetLastError() to obtain error 
  // information.
  //
  // NOTE: The return value FALSE indicates an irrecoverably bad state
  // of the process, usually caused by a bad API call such as entering 
  // or leaving a destroyed critical section.
  BOOL LeaveRead();

  // Acquires the write lock. The first argument is the timeout in 
  // milliseconds. The alertable flag must be true if the wait should 
  // complete when a user APC is queued.
  //
  // In case of success, the return value is WAIT_OBJECT_0. See
  // the documentation of the Win32 function SignalObjectAndWait()
  // for an explanation of the other possible return values.
  //
  // NOTE: The return value WAIT_FAILED indicates an irrecoverably bad
  // state of the process, usually caused by a bad API call such as 
  // entering or leaving a destroyed critical section.
  DWORD EnterWrite(
    DWORD dwMillisecondsTimeout = INFINITE,
    BOOL bAlertable = FALSE
    );

  // Releases the write lock. In case of success,  the return value is
  // TRUE. In case of failure, call GetLastError() to obtain error 
  // information.
  //
  // NOTE: The return value FALSE indicates an irrecoverably bad state
  // of the process, usually caused by a bad API call such as entering 
  // or leaving a destroyed critical section.
  BOOL LeaveWrite();

  // Checks if the handles to the synch objects have been
  // created.
  bool IsValid();

private:

  // Private helper functions
  // ========================
  
  // Data members
  // ============

  // Read/write monitor
  MultiSetMonitor m_monReadWrite;

  // Nunmber of currently active readers.
  int m_nNumActiveReaders;

  // Number of currently waiting writers.
  int m_nNumWaitingWriters;

  // Flag to indicate an active writer.
  bool m_bWriteInProgress;

};

#endif
//
// End multiple inclusion protection
